'''
Test cases for docker container creation.
'''

#adding for py3 support
from __future__ import absolute_import

import logging
import os
import random
import re
import shutil
import tempfile

import unittest

import pexpect


# Configure logging. Troubleshooting the pexpect interactions in
# particular needs a lot of tracing.
FILENAME = os.path.split(__file__)[-1]
shandler = logging.StreamHandler()
sformatter = logging.Formatter(
    '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )
shandler.setFormatter(sformatter)
logger = logging.getLogger()
logger.setLevel(logging.WARNING)
# logger.setLevel(logging.INFO)
# logger.setLevel(logging.DEBUG)
logger.addHandler(shandler)

# These need to be removed from the test run environment if present
# before configuring the environment for the pexpect call to the make
# process.
BUILD_CONTROL_ENV_VARIABLES = [
    'AARDVARK_ROLE',
    'AARDVARK_DB_URI',
    'SWAG_BUCKET',
    'AARDVARK_IMAGES_TAG'
    ]

# We'll copy the docker directory contents to the temporary working
# directory each time.
DOCKER_PATH = os.path.join(
    os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
    "docker"
    )

# An index of possible docker images and their pseudo-artifacts.
# TODO: Preserved for reference purposes; the DOCKER_IMAGES variable
# isn't used as of this commenting.
DOCKER_IMAGES = {
    'aardvark-base': 'aardvark-base-docker-build',
    'aardvark-data-volume': 'aardvark-data-docker-build',
    'aardvark-data-volume': 'aardvark-data-docker-run',
    'aardvark-apiserver': 'aardvark-apiserver-docker-build',
    'aardvark-collector': 'aardvark-collector-docker-build',
    }

# The subdirectory of the working directory where pseudo-artifacts
# are created.
ARTIFACT_DIRECTORY = 'artifacts'

# A few constants that are checked when testing container
# config settings.
CONTAINER_CONFIG_PATH = '/etc/aardvark/config.py'

EXPECTED_SQLITE_DB_URI = 'sqlite:////usr/share/aardvark-data/aardvark.db'
EXPECTED_SQL_TRACK_MODS = False

# Making targets can take some time, and depends on network connection
# speed. Set the NETWORK_SPEED_FACTOR environment variable to increase
# if necessary.
PEXPECT_TIMEOUTS = {
    'default': 30,
    'container_command': 2,
    'aardvark': 30,
    'aardvark-all': 240,
    'aardvark-base': 180,
    'aardvark-sqlite': 300,
    }
NETWORK_SPEED_FACTOR = 1.0

# The key that will uniquely identify a docker construct of the
# indicated type.
UID_KEY = {
    'image': 'id',
    'container': 'id',
    'volume': 'name'
    }

# A message for unittest.skipIf.
SKIP_NO_IMAGES_TAG_MSG = (
    "Remove aardvark-*:latest images and the aardvark-data volume as desired"
    " and set the RUN_AARDVARK_DOCKER_TESTS_NO_IMAGES_TAG environment variable"
    " to 1 to enable tests that don't set the AARDVARK_IMAGES_TAG."
    )


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def sqlite_db_uri(path):
    '''Return the default db_uri value at runtime.'''
    return 'sqlite:///{}/aardvark.db'.format(path)


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def copy_recipes(src, dest):
    '''Copy Make- and Dockerfiles from src to dest'''
    [
        shutil.copyfile(os.path.join(src, f), os.path.join(dest, f))
        for f in os.listdir(src)
        if (f.startswith("Dockerfile") or f.startswith("Makefile"))
        ]


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def interact(
        spawn,
        command,
        response_filter=None,
        timeout=PEXPECT_TIMEOUTS['container_command']
        ):  # pylint: disable=bad-continuation
    '''Send command to spawn and return filtered response.'''

    def default_response_filter(response):
        '''Define a default method for the response filter.'''
        return response

    if not response_filter:
        response_filter = default_response_filter

    shell_prompt = r'root@\w+:[\S]+# '
    expect_prompt = [shell_prompt, pexpect.EOF, pexpect.TIMEOUT]

    eof_position = expect_prompt.index(pexpect.EOF)
    timeout_position = expect_prompt.index(pexpect.TIMEOUT)

    eof_msg_template = 'unexpected EOF in pexpect connection to {}'
    responses = []
    result = None

    logger.info('COMMAND:\t%s', command)
    spawn.sendline(command)

    # Read until we time out, indicating no more lines are pending;
    # watch for the response that has our command echoed on one line
    # and our response on the next.
    while(True):

        prompt_index = spawn.expect(expect_prompt, timeout=timeout)

        import json
        if prompt_index < timeout_position:
            responses.append(
                spawn.before.decode('utf-8').replace('\r\n', '\n')
                )
            # print '--------'
            logger.debug('    BEFORE: %s', json.dumps(responses[-1]))
            logger.debug(
                '    AFTER: %s',
                json.dumps(spawn.after.decode("utf-8").replace('\r\n', '\n'))
                )
            if command and responses[-1].startswith(command + '\n'):
                result = response_filter(
                    responses[-1].replace(command + '\n', '', 1).strip()
                    )
                logger.info("    result found: %s", result)
            # print '--------\n'

        elif prompt_index == timeout_position:
            logger.debug('    TIMEOUT')
            break

        elif prompt_index == eof_position:
            raise RuntimeError(eof_msg_template.format(command))

    logger.debug('Responses:\n%s', '\n'.join(responses))
    logger.debug('')

    return result


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TestDockerBase(unittest.TestCase):
    '''Base class for docker container construction test cases.'''

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @classmethod
    def setUpClass(cls):
        '''Test case class common fixture setup.'''

        # These will record the constructs at class start; we'll log
        # any discrepancy at class termination.
        cls.constructs = {'original': {}, 'current': {}}

        for construct_type in ['image', 'container', 'volume']:
            cls.constructs['original'][construct_type] = (
                cls.get_docker_constructs(construct_type)
                )
            cls.constructs['current'][construct_type] = list(
                cls.constructs['original'][construct_type]
                )

        cls.tmpdir = tempfile.mkdtemp()

        cls.original_working_dir = os.getcwd()
        os.chdir(cls.tmpdir)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @classmethod
    def tearDownClass(cls):
        '''Test case class common fixture teardown.'''

        os.chdir(cls.original_working_dir)
        cls.clean_tmpdir()
        os.rmdir(cls.tmpdir)

        cls.warn_on_hanging_constructs()

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @classmethod
    def warn_on_hanging_constructs(cls):
        '''Log warnings for undeleted docker constructs created by tests.'''

        construct_fields = {
            'container': ['id', 'name'],
            'image': ['id', 'name', 'tag'],
            'volume': ['name'],
            }

        hanging_constructs = {}

        for construct_type in construct_fields.keys():

            hanging_constructs[construct_type] = [
                c for c in cls.constructs['current'][construct_type]
                if c[UID_KEY[construct_type]] not in [
                    x[UID_KEY[construct_type]]
                    for x in cls.constructs['original'][construct_type]
                    ]
                ]

            for construct in hanging_constructs[construct_type]:
                logger.warning(
                    'a failed test case left behind %s:\t%s',
                    construct_type,
                    '\t'.join([
                        construct[field]
                        for field in construct_fields[construct_type]
                        ])
                    )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @classmethod
    def clean_tmpdir(cls):
        '''Remove all content from cls.tmpdir.'''

        for root, dirs, files in os.walk(cls.tmpdir, topdown=False):

            for name in files:

                path = os.path.join(root, name)
                os.remove(path)

            for name in dirs:

                path = os.path.join(root, name)
                if os.path.islink(path):
                    os.remove(path)
                else:
                    os.rmdir(path)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def setUp(self):
        '''Test case common fixture setup.'''

        self.clean_tmpdir()

        # This copies the Makefile and the Dockerfiles.
        copy_recipes(DOCKER_PATH, self.tmpdir)

        # We track to make sure we don't disturb any files that were
        # present in the working directory.
        self.initial_contents = os.listdir(self.tmpdir)

        # If true, we will stop containers and delete images created
        # during each test case.
        self.delete_artifacts = True

        # A (almost certainly) unique string for unique test case
        # artifact names.
        self.testcase_tag = (
            'test{:08X}'.format(random.randrange(16 ** 8)).lower()
            )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def tearDown(self):
        '''Test case common fixture teardown.'''

        logger.info("======= tearDown: %s", self.delete_artifacts)

        self.log_docker_constructs()

        new_containers = self.new_constructs('container')
        new_images = self.new_constructs('image')
        new_volumes = self.new_constructs('volume')

        if self.delete_artifacts:
            # Every case should clean up its docker images and containers.

            for container in new_containers:
                # These should have been launched with the --rm flag,
                # so they should be removed once stopped.
                logger.info("REMOVING %s", container['id'])
                pexpect.run('docker stop {}'.format(container['id']))

            for image in new_images:
                logger.info("REMOVING %s", image['id'])
                pexpect.run('docker rmi {}'.format(image['id']))

            for volume in new_volumes:
                logger.info("REMOVING %s", volume['name'])
                pexpect.run('docker volume rm {}'.format(volume['name']))

        else:
            # We'll leave behind any new docker constructs, so we need
            # to update the "original docker volumes".
            self.constructs['current']['container'].extend(new_containers)
            self.constructs['current']['image'].extend(new_images)
            self.constructs['current']['volume'].extend(new_volumes)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def log_docker_constructs(self, **kwargs):
        '''Log docker construct status.'''

        def log_listing(caption, records, *fields):
            '''Helper method for log message construction.'''

            return '{}:\n    '.format(caption) + '\n    '.join([
                '\t'.join(
                    [record[field] for field in fields]
                    ) for record in records
                ])

        construct_fields = {
            'container': ['id', 'name'],
            'image': ['id', 'name', 'tag'],
            'volume': ['name'],
            }

        for construct_type in construct_fields.keys():

            logger.info("------------%ss:", construct_type)
            logger.info(log_listing(
                'Original',
                self.constructs['original'][construct_type],
                *construct_fields[construct_type]
                ))
            logger.info(log_listing(
                'Current',
                self.get_docker_constructs(construct_type),
                *construct_fields[construct_type]
                ))
            logger.info(log_listing(
                'New',
                self.new_constructs(construct_type),
                *construct_fields[construct_type]
                ))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @classmethod
    def get_docker_constructs(cls, construct_type, *expected_constructs):
        '''Get a list of images, containers or volumes.'''
        return {
            'image': cls.get_docker_image_list,
            'container': cls.get_docker_container_list,
            'volume': cls.get_docker_volume_list
            }[construct_type](*expected_constructs)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @staticmethod
    def get_docker_image_list(*expected_images):
        '''Get the output from "docker image ls" for specified image names.'''

        image_listing_pattern = (
            r'(?P<name>[^\s]+)\s+'
            r'(?P<tag>[^\s]+)\s+'
            r'(?P<id>[0-9a-f]+)\s+'
            r'(?P<created>.+ago)\s+'
            r'(?P<size>[^\s]+)'
            r'\s*$'
            )
        image_listing_re = re.compile(image_listing_pattern)

        docker_images_response = pexpect.run('docker image ls')

        image_list = []
        expected_image_nametag_pairs = [
            (x.split(':') + ['latest'])[0:2] for x in expected_images
            ] if expected_images else None

        docker_images_response_l = docker_images_response.decode('utf-8').split('\n')

        for line in docker_images_response_l:
            match = image_listing_re.match(line)
            if (
                    match and (
                        not expected_images or [
                            match.groupdict()['name'], match.groupdict()['tag']
                            ] in expected_image_nametag_pairs
                        )
                    ):
                image_list.append(match.groupdict())

        return image_list

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @staticmethod
    def get_docker_container_list(*expected_containers):
        '''Get the output from "docker ps -a" for specified container names.'''

        container_listing_pattern = (
            r'(?P<id>[0-9a-f]+)\s+'
            r'(?P<image>[^\s]+)\s+'
            r'(?P<command>"[^"]+")\s+'
            r'(?P<created>.+ago)\s+'
            r'(?P<status>(Created|Exited.*ago|Up \d+ \S+))\s+'
            r'(?P<ports>[^\s]+)?\s+'
            r'(?P<name>[a-z]+_[a-z]+)'
            # r'\s*$'
            )
        container_listing_re = re.compile(container_listing_pattern)

        docker_containers_response = pexpect.run('docker ps -a')

        container_list = []
        # expected_container_nametag_pairs = [
        #     (x.split(':') + ['latest'])[0:2] for x in expected_containers
        #     ] if expected_containers else []

        docker_containers_response_l = docker_containers_response.decode('utf-8').split('\n')

        for line in docker_containers_response_l:
            match = container_listing_re.match(line)
            if match:
                container_list.append(match.groupdict())

        return container_list

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @staticmethod
    def get_docker_volume_list(*expected_volumes):
        '''Get the output from "docker volume ls" for specified volumes.'''

        volume_listing_pattern = (
            r'(?P<driver>\S+)\s+'
            r'(?P<name>\S+)'
            # r'\s*$'
            )
        volume_listing_re = re.compile(volume_listing_pattern)

        docker_volumes_response = pexpect.run('docker volume ls')

        docker_volumes_response_l = docker_volumes_response.decode('utf-8').split('\n')

        volume_list = []

        for line in docker_volumes_response_l:
            match = volume_listing_re.match(line)
            if match:
                volume_list.append(match.groupdict())

        return volume_list

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def require_filenames_in_directory(self, patterns=None, directory='.'):
        '''Check that filenames are found in the indicated directory.

        Each pattern in the list of patterns must match exactly one
        file in the indicated directory.

        '''

        failure_string_template = (
            'Unexpected or missing filename match result in {}'
            ' for pattern r\'{}\':\n{}\n'
            'Directory contents:\n{}'
            )

        if patterns:
            self.assertTrue(os.path.exists(directory))
            all_filenames = os.listdir(directory)
            for pattern in patterns:
                matching_files = [
                    x for x in all_filenames
                    if re.match(pattern, x)
                    ]
                self.assertTrue(
                    len(matching_files) == 1,
                    failure_string_template.format(
                        directory,
                        pattern,
                        matching_files,
                        os.listdir(directory)
                        )
                    )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def new_constructs(self, construct_type):
        '''Get a list of images, containers or volumes not already recorded.'''
        return [
            c for c in self.get_docker_constructs(construct_type)
            if c[UID_KEY[construct_type]] not in [
                o[UID_KEY[construct_type]]
                for o in self.constructs['original'][construct_type]
                ]
            ]

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def get_container_details(self, image):
        '''Run a docker container shell and retrieve several details.'''

        # (detail name, command, result filter) for extracting details
        # from container command lines.
        shell_commands = (
            ('pwd', 'pwd', None),
            ('config_file', 'ls {}'.format(CONTAINER_CONFIG_PATH), None),
            ('config_contents', 'cat {}'.format(CONTAINER_CONFIG_PATH), None),
            )

        command = "docker run --rm -it {} bash".format(image)
        logger.info('IMAGE: %s', image)
        logger.info('CONTAINER LAUNCH COMMAND: %s', command)
        spawn = pexpect.spawn(command)

        container_details = {}

        for field, shell_command, response_filter in shell_commands:
            container_details[field] = interact(
                spawn, shell_command, response_filter
                )

        # Exit the container.
        spawn.sendcontrol('d')

        # "Expand" the config records if we found a config file.
        if container_details['config_file'] == CONTAINER_CONFIG_PATH:
            try:
                exec(container_details['config_contents'], container_details)
            except SyntaxError:
                pass
            # The '__builtins__' are noise:
            if '__builtins__' in container_details:
                del container_details['__builtins__']

        return container_details

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def check_container_details(self, image, expected_details=None):
        '''Validate docker container details.'''

        # A helper method to retrieve container details of interest,
        # necessary for configuration items that aren't simple config
        # file values; e.g. "SWAG_BUCKET".
        # As written, this won't allow tests to catch *missing*
        # entries if the *expected* value is None.
        def get_detail_value(data, detail):
            '''A helper method to retrieve container details of interest.'''

            def default_get_detail_value_method(data):
                '''Define a default method for get_detail_value.'''
                return data.get(detail)

            method = {
                'SWAG_BUCKET': lambda data: (
                    data.get('SWAG_OPTS') or {}
                    ).get('swag.bucket_name')
                }.get(detail)

            if method is None:
                method = default_get_detail_value_method

            return method(data)

        image_name, image_tag = image.split(':')

        expected_details = expected_details or {}
        assert '_common' in expected_details

        expected_container_details = dict(expected_details['_common'])
        if image_name in expected_details:
            expected_container_details.update(expected_details[image_name])

        container_details = self.get_container_details(image)

        clean_comparison = {
            'image': image,
            'missing': {},
            'incorrect': {}
            }
        comparison = {
            'image': image,
            'missing': {},
            'incorrect': {}
            }

        for k, v in expected_container_details.items():

            logger.info(' -- checking configuration item %s...', k)
            logger.info('    expected: %s', v)

            actual = get_detail_value(container_details, k)

            # if k not in container_details:
            # TODO: Note that this fails if we're *expecting* None.
            if actual is None:
                comparison['missing'][k] = v
                logger.info('    actual: -')

            elif actual != v:
                comparison['incorrect'][k] = {
                    'expected': v,
                    'actual': actual
                    }
                logger.info(comparison['incorrect'][k])

            else:
                logger.info('    actual:   %s', actual)

        logger.info("comparing %s", image)
        logger.info(comparison)
        logger.info('')
        self.assertEqual(comparison, clean_comparison)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def case_worker(
            self,
            target,
            expected_artifacts=None,
            expected_docker_images=None,
            expected_details=None,
            expect_aardvark=True,
            add_env=None,
            set_images_tag=True,
            ):
        '''Carry out common test steps.
        '''

        logger.info(' -' * 8 + ' working case: %s' + ' -' * 8, target)

        # Unless we finish without a failure Exception, tell tearDown
        # not to clean up artifacts. We reset this below.
        self.delete_artifacts = False

        # A unique string to add to certain test case artifact names
        # to avoid clobbering/colliding.
        if set_images_tag:
            images_tag = self.testcase_tag
            logger.info('Test case images tag is %s', images_tag)
        else:
            images_tag = 'latest'
            logger.info('Default test case images tag will be "latest"')

        expected_artifacts = expected_artifacts or []
        expected_docker_images = expected_docker_images or []
        tagged_expected_docker_images = [
            x + ':{}'.format(images_tag)
            for x in expected_docker_images
            ]

        # expected_details is a two level dict so a straightforward
        # dict update isn't possible.
        expected_details = expected_details or {}
        expected_details['_common'] = expected_details.get('_common') or {}
        expected_details['_common']['config_file'] = CONTAINER_CONFIG_PATH

        # Environment variables to add to the pexpect interaction with
        # containers.
        add_env = dict(add_env or {})
        if set_images_tag:
            add_env = dict(add_env, AARDVARK_IMAGES_TAG=images_tag)

        # Fetch the default environment settings so we can update
        # those with specific case settings.
        spawn_env = dict(
            map(
                lambda x: x.strip().split('=', 1),
                pexpect.run('env').strip().decode("utf-8").split("\n")
                )
            )

        # Remove any build control variables we inherit from the test
        # environment - we want complete control over which are
        # visible to the make process in the pexpect call.
        spawn_env = {
            k: v for (k, v) in spawn_env.items()
            if k not in BUILD_CONTROL_ENV_VARIABLES
        }

        command = 'make {}'.format(target)
        logger.info('COMMAND: %s', command)

        # TODO: A sort of halfhearted attempt at adjusting for network
        # conditions. Need some kind of not-too-slow way to check for
        # network speed, say a sample download or something.
        (result, exitstatus) = pexpect.run(
            command,
            timeout=(
                PEXPECT_TIMEOUTS.get(target) or PEXPECT_TIMEOUTS['default'] *
                NETWORK_SPEED_FACTOR
                ),
            withexitstatus=True,
            env=dict(spawn_env, **add_env)
            )

        self.assertEqual(
            exitstatus, 0,
            'command "{}" exited with exit status {}'.format(
                command, exitstatus
                )
            )

        # Sanity check - we didn't delete the Makefile or any of the
        # Dockerfiles.
        self.assertEqual(
            [x for x in self.initial_contents if x not in os.listdir('.')],
            []
            )

        if expected_docker_images:
            self.assertCountEqual(
                [
                    [x['name'], x['tag']]
                    for x in self.get_docker_image_list(
                        *tagged_expected_docker_images
                        )
                    ],
                [x.split(':') for x in tagged_expected_docker_images]
                )

        for image in tagged_expected_docker_images:
            self.check_container_details(image, expected_details)

        if expect_aardvark:
            self.require_filenames_in_directory([r'aardvark$'])

        if expected_artifacts:
            self.require_filenames_in_directory(
                expected_artifacts,
                directory=ARTIFACT_DIRECTORY
                )

        # We made it through, tell tearDown we can clean up artifacts.
        self.delete_artifacts = True


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TestDockerContainerConstruction(TestDockerBase):
    '''Test cases for docker container construction.'''

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def test_make_aardvark(self):
        '''Test "make aardvark".'''

        self.case_worker(
            target='aardvark',
            expect_aardvark=True,
            )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def test_make_aardvark_base(self):
        '''Test "make aardvark-base".'''

        self.case_worker(
            target='aardvark-base',
            expected_docker_images=[
                'aardvark-base'
                ],
            expected_details={
                '_common': {
                    'pwd': '/etc/aardvark',
                    'NUM_THREADS': 5,
                    'ROLENAME': 'Aardvark',
                    'SQLALCHEMY_DATABASE_URI': EXPECTED_SQLITE_DB_URI,
                    'SQLALCHEMY_TRACK_MODIFICATIONS': EXPECTED_SQL_TRACK_MODS,
                    }
                },
            expected_artifacts=[
                'aardvark-base-docker-build'
                ],
            )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def test_make_aardvark_base_set_env_variables(self):
        '''Test "make aardvark-base" with build time environment variables.'''

        aardvark_db_uri = 'blort://blah.bleh.bloo/bing/bang/bong'
        aardvark_role = 'llama'
        swag_bucket = 'ponponpon'

        self.case_worker(
            target='aardvark-base',
            expected_docker_images=[
                'aardvark-base'
                ],
            expected_details={
                '_common': {
                    'pwd': '/etc/aardvark',
                    'NUM_THREADS': 5,
                    'ROLENAME': aardvark_role,
                    'SQLALCHEMY_DATABASE_URI': aardvark_db_uri,
                    'SQLALCHEMY_TRACK_MODIFICATIONS': EXPECTED_SQL_TRACK_MODS,
                    'SWAG_BUCKET': swag_bucket,
                    }
                },
            expected_artifacts=[
                'aardvark-base-docker-build'
                ],
            add_env={
                'AARDVARK_DB_URI': aardvark_db_uri,
                'AARDVARK_ROLE': aardvark_role,
                'SWAG_BUCKET': swag_bucket,
                }
            )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def test_make_aardvark_all(self):
        '''Test "make aardvark-all".'''

        self.case_worker(
            target='aardvark-all',
            expected_docker_images=[
                'aardvark-base',
                'aardvark-collector',
                'aardvark-apiserver',
                ],
            expected_details={
                '_common': {
                    'pwd': '/usr/share/aardvark-data',
                    'NUM_THREADS': 5,
                    'ROLENAME': 'Aardvark',
                    'SQLALCHEMY_DATABASE_URI': EXPECTED_SQLITE_DB_URI,
                    'SQLALCHEMY_TRACK_MODIFICATIONS': EXPECTED_SQL_TRACK_MODS,
                    },
                'aardvark-base': {
                    'pwd': '/etc/aardvark',
                    },
                },
            expected_artifacts=[
                'aardvark-base-docker-build',
                'aardvark-apiserver-docker-build',
                'aardvark-collector-docker-build',
                ],
            )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @unittest.skipIf(
        not os.environ.get('RUN_AARDVARK_DOCKER_TESTS_NO_IMAGES_TAG'),
        SKIP_NO_IMAGES_TAG_MSG
        )
    def test_make_aardvark_all_no_images_tag(self):
        '''Test "make aardvark-all" without specifying the images tag.'''

        self.case_worker(
            target='aardvark-all',
            expected_docker_images=[
                'aardvark-base',
                'aardvark-collector',
                'aardvark-apiserver',
                ],
            expected_details={
                '_common': {
                    'pwd': '/usr/share/aardvark-data',
                    'NUM_THREADS': 5,
                    'ROLENAME': 'Aardvark',
                    'SQLALCHEMY_DATABASE_URI': EXPECTED_SQLITE_DB_URI,
                    'SQLALCHEMY_TRACK_MODIFICATIONS': EXPECTED_SQL_TRACK_MODS,
                    },
                'aardvark-base': {
                    'pwd': '/etc/aardvark',
                    },
                },
            expected_artifacts=[
                'aardvark-base-docker-build',
                'aardvark-apiserver-docker-build',
                'aardvark-collector-docker-build',
                ],
            set_images_tag=False
            )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def test_make_aardvark_sqlite(self):
        '''Test "make aardvark-sqlite".'''

        self.case_worker(
            target='aardvark-sqlite',
            expected_docker_images=[
                'aardvark-base',
                'aardvark-data-init',
                'aardvark-collector',
                'aardvark-apiserver',
                ],
            expected_details={
                '_common': {
                    'pwd': '/usr/share/aardvark-data',
                    'NUM_THREADS': 5,
                    'ROLENAME': 'Aardvark',
                    'SQLALCHEMY_DATABASE_URI': EXPECTED_SQLITE_DB_URI,
                    'SQLALCHEMY_TRACK_MODIFICATIONS': EXPECTED_SQL_TRACK_MODS,
                    },
                'aardvark-base': {
                    'pwd': '/etc/aardvark',
                    },
                },
            expected_artifacts=[
                'aardvark-base-docker-build',
                'aardvark-data-docker-build',
                'aardvark-data-docker-run',
                'aardvark-apiserver-docker-build',
                'aardvark-collector-docker-build',
                ],
            )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @unittest.skipIf(
        not os.environ.get('RUN_AARDVARK_DOCKER_TESTS_NO_IMAGES_TAG'),
        SKIP_NO_IMAGES_TAG_MSG
        )
    def test_make_aardvark_sqlite_no_images_tag(self):
        '''Test "make aardvark-sqlite" without specifying the images tag.'''

        self.case_worker(
            target='aardvark-sqlite',
            expected_docker_images=[
                'aardvark-base',
                'aardvark-data-init',
                'aardvark-collector',
                'aardvark-apiserver',
                ],
            expected_details={
                '_common': {
                    'pwd': '/usr/share/aardvark-data',
                    'NUM_THREADS': 5,
                    'ROLENAME': 'Aardvark',
                    'SQLALCHEMY_DATABASE_URI': EXPECTED_SQLITE_DB_URI,
                    'SQLALCHEMY_TRACK_MODIFICATIONS': EXPECTED_SQL_TRACK_MODS,
                    },
                'aardvark-base': {
                    'pwd': '/etc/aardvark',
                    },
                },
            expected_artifacts=[
                'aardvark-base-docker-build',
                'aardvark-data-docker-build',
                'aardvark-data-docker-run',
                'aardvark-apiserver-docker-build',
                'aardvark-collector-docker-build',
                ],
            set_images_tag=False
            )


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Define test suites.
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
load_case = unittest.TestLoader().loadTestsFromTestCase
all_suites = {
    'testdockercontainerconstruction': load_case(
        TestDockerContainerConstruction
        ),
    }

master_suite = unittest.TestSuite(all_suites.values())

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
if __name__ == '__main__':
    unittest.main()



#ensure absolute import for python3
from __future__ import absolute_import

import os
try:
    import queue as Queue # Queue renamed to queue in py3
except ModuleNotFoundError:
    import Queue
import re
import threading

import better_exceptions # noqa
from blinker import Signal
from bunch import Bunch
from flask import current_app
from flask_script import Manager, Command, Option
from swag_client.backend import SWAGManager
from swag_client.exceptions import InvalidSWAGDataException
from swag_client.util import parse_swag_config_options

from aardvark import create_app, db
from aardvark.updater import AccountToUpdate

try:               # Python 2
    raw_input
except NameError:  # Python 3
    raw_input = input

try:               # Python 2
    unicode
except NameError:  # Python 3
    unicode = str

manager = Manager(create_app)

ACCOUNT_QUEUE = Queue.Queue()
DB_LOCK = threading.Lock()
QUEUE_LOCK = threading.Lock()
UPDATE_DONE = False

SWAG_REPO_URL = 'https://github.com/Netflix-Skunkworks/swag-client'

LOCALDB = 'sqlite'

# Configuration default values.
DEFAULT_LOCALDB_FILENAME = 'aardvark.db'
DEFAULT_SWAG_BUCKET = 'swag-data'
DEFAULT_AARDVARK_ROLE = 'Aardvark'
DEFAULT_NUM_THREADS = 5  # testing shows problems with more than 6 threads


class UpdateAccountThread(threading.Thread):
    global ACCOUNT_QUEUE, DB_LOCK, QUEUE_LOCK, UPDATE_DONE
    on_ready = Signal()
    on_complete = Signal()
    on_failure = Signal()

    def __init__(self, thread_ID):
        self.thread_ID = thread_ID
        threading.Thread.__init__(self)
        self.app = current_app._get_current_object()

    def run(self):
        while not UPDATE_DONE:
            self.on_ready.send(self)

            QUEUE_LOCK.acquire()

            if not ACCOUNT_QUEUE.empty():
                (account_num, role_name, arns) = ACCOUNT_QUEUE.get()

                self.app.logger.info("Thread #{} updating account {} with {} arns".format(
                                     self.thread_ID, account_num, 'all' if arns[0] == 'all' else len(arns)))

                QUEUE_LOCK.release()

                account = AccountToUpdate(self.app, account_num, role_name, arns)
                ret_code, aa_data = account.update_account()

                if ret_code != 0:  # retrieve wasn't successful, put back on queue
                    self.on_failure.send(self)
                    QUEUE_LOCK.acquire()
                    ACCOUNT_QUEUE.put((account_num, role_name, arns))
                    QUEUE_LOCK.release()

                self.app.logger.info("Thread #{} persisting data for account {}".format(self.thread_ID, account_num))

                DB_LOCK.acquire()
                persist_aa_data(self.app, aa_data)
                DB_LOCK.release()

                self.on_complete.send(self)
                self.app.logger.info("Thread #{} FINISHED persisting data for account {}".format(self.thread_ID, account_num))
            else:
                QUEUE_LOCK.release()


def persist_aa_data(app, aa_data):
    """
    Reads access advisor JSON file & persists to our database
    """
    from aardvark.model import AWSIAMObject, AdvisorData

    with app.app_context():
        if not aa_data:
            app.logger.warn('Cannot persist Access Advisor Data as no data was collected.')
            return

        arn_cache = {}
        for arn, data in aa_data.items():
            if arn in arn_cache:
                item = arn_cache[arn]
            else:
                item = AWSIAMObject.get_or_create(arn)
                arn_cache[arn] = item
            for service in data:
                AdvisorData.create_or_update(item.id,
                                             service['LastAuthenticated'],
                                             service['ServiceName'],
                                             service['ServiceNamespace'],
                                             service.get('LastAuthenticatedEntity'),
                                             service['TotalAuthenticatedEntities'])
        db.session.commit()


@manager.command
def drop_db():
    """ Drops the database. """
    db.drop_all()


@manager.command
def create_db():
    """ Creates the database. """
    db.create_all()


# All of these default to None rather than the corresponding DEFAULT_* values
# so we can tell whether they were passed or not. We don't prompt for any of
# the options that were passed as parameters.
@manager.option('-a', '--aardvark-role', dest='aardvark_role_param', type=unicode)
@manager.option('-b', '--swag-bucket', dest='bucket_param', type=unicode)
@manager.option('-d', '--db-uri', dest='db_uri_param', type=unicode)
@manager.option('--num-threads', dest='num_threads_param', type=int)
@manager.option('--no-prompt', dest='no_prompt', action='store_true', default=False)
def config(aardvark_role_param, bucket_param, db_uri_param, num_threads_param, no_prompt):
    """
    Creates a config.py configuration file from user input or default values.

    If all configurable values are specified by parameters, user input
    is not needed and will not be prompted.

    If the no-prompt flag is not set, user input will be prompted for
    each of the configurable values not specified by parameters.

    If the no-prompt flag is set, no user input will be collected and
    the configuration file will be populated with option-specified values
    or defaults.

    The resulting configuration file defines the following parameters.
    Configurable parameters are shown in <angle braces>.

    SWAG_OPTS = {'swag.type': 's3', 'swag.bucket_name': <bucket>}
    SWAG_FILTER = None
    SWAG_SERVICE_ENABLED_REQUIREMENT = None
    ROLENAME = <aardvark_role>
    REGION = "us-east-1"
    SQLALCHEMY_DATABASE_URI = <db_uri>
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    NUM_THREADS = <num_threads>
    LOG_CFG = {...}
    """
    # We don't set these until runtime.
    default_db_uri = '{localdb}:///{path}/{filename}'.format(
        localdb=LOCALDB, path=os.getcwd(), filename=DEFAULT_LOCALDB_FILENAME
        )

    if no_prompt:  # Just take the parameters as currently constituted.
        aardvark_role = aardvark_role_param or DEFAULT_AARDVARK_ROLE
        num_threads = num_threads_param or DEFAULT_NUM_THREADS
        db_uri = db_uri_param or default_db_uri

        # If a swag bucket was specified we set write_swag here so it gets
        # written out to the config file below.
        write_swag = bool(bucket_param)
        bucket = bucket_param or DEFAULT_SWAG_BUCKET

    else:
        # This is essentially the same "param, or input, or default"
        # structure as the additional parameters below.
        if bucket_param:
            bucket = bucket_param
            write_swag = True
        else:
            print('\nAardvark can use SWAG to look up accounts. See {repo_url}'.format(repo_url=SWAG_REPO_URL))
            use_swag = raw_input('Do you use SWAG to track accounts? [yN]: ')
            if len(use_swag) > 0 and 'yes'.startswith(use_swag.lower()):
                bucket_prompt = 'SWAG_BUCKET [{default}]: '.format(default=DEFAULT_SWAG_BUCKET)
                bucket = raw_input(bucket_prompt) or DEFAULT_SWAG_BUCKET
                write_swag = True
            else:
                write_swag = False

        aardvark_role_prompt = 'ROLENAME [{default}]: '.format(default=DEFAULT_AARDVARK_ROLE)
        db_uri_prompt = 'DATABASE URI [{default}]: '.format(default=default_db_uri)
        num_threads_prompt = '# THREADS [{default}]: '.format(default=DEFAULT_NUM_THREADS)

        aardvark_role = aardvark_role_param or raw_input(aardvark_role_prompt) or DEFAULT_AARDVARK_ROLE
        db_uri = db_uri_param or raw_input(db_uri_prompt) or default_db_uri
        num_threads = num_threads_param or raw_input(num_threads_prompt) or DEFAULT_NUM_THREADS

    log = """LOG_CFG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s %(levelname)s: %(message)s '
                '[in %(pathname)s:%(lineno)d]'
        }
    },
    'handlers': {
        'file': {
            'class': 'logging.handlers.RotatingFileHandler',
            'level': 'DEBUG',
            'formatter': 'standard',
            'filename': 'aardvark.log',
            'maxBytes': 10485760,
            'backupCount': 100,
            'encoding': 'utf8'
        },
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'standard',
            'stream': 'ext://sys.stdout'
        }
    },
    'loggers': {
        'aardvark': {
            'handlers': ['file', 'console'],
            'level': 'DEBUG'
        }
    }
}"""

    with open('config.py', 'w') as filedata:
        print('\n>> Writing to config.py')
        filedata.write('# Autogenerated config file\n')
        if write_swag:
            filedata.write("SWAG_OPTS = {{'swag.type': 's3', 'swag.bucket_name': '{bucket}'}}\n".format(bucket=bucket))
            filedata.write("SWAG_FILTER = None\n")
            filedata.write("SWAG_SERVICE_ENABLED_REQUIREMENT = None\n")
        filedata.write('ROLENAME = "{role}"\n'.format(role=aardvark_role))
        filedata.write('REGION = "us-east-1"\n')
        filedata.write('ARN_PARTITION = "aws"\n')
        filedata.write('SQLALCHEMY_DATABASE_URI = "{uri}"\n'.format(uri=db_uri))
        filedata.write('SQLALCHEMY_TRACK_MODIFICATIONS = False\n')
        filedata.write('NUM_THREADS = {num_threads}\n'.format(num_threads=num_threads))
        filedata.write(log)


@manager.option('-a', '--accounts', dest='accounts', type=unicode, default='all')
@manager.option('-r', '--arns', dest='arns', type=unicode, default='all')
def update(accounts, arns):
    """
    Asks AWS for new Access Advisor information.
    """
    accounts = _prep_accounts(accounts)
    arns = arns.split(',')
    app = create_app()

    global ACCOUNT_QUEUE, QUEUE_LOCK, UPDATE_DONE

    role_name = app.config.get('ROLENAME')
    num_threads = app.config.get('NUM_THREADS') or 5

    if num_threads > 6:
        current_app.logger.warn('Greater than 6 threads seems to cause problems')

    QUEUE_LOCK.acquire()
    for account_number in accounts:
        ACCOUNT_QUEUE.put((account_number, role_name, arns))
    QUEUE_LOCK.release()

    threads = []
    for thread_num in range(num_threads):
        thread = UpdateAccountThread(thread_num + 1)
        thread.start()
        threads.append(thread)

    while not ACCOUNT_QUEUE.empty():
        pass
    UPDATE_DONE = True


def _prep_accounts(account_names):
    """
    Convert CLI provided account names into list of accounts from SWAG.
    Considers account aliases as well as account names.
    Returns a list of account numbers
    """
    matching_accounts = list()
    account_names = account_names.split(',')
    account_names = {name.lower().strip() for name in account_names}

    # create a new copy of the account_names list so we can remove accounts as needed
    for account in list(account_names):
        if re.match('\d{12}', account):
            account_names.remove(account)
            matching_accounts.append(account)

    if not account_names:
        return matching_accounts

    try:
        current_app.logger.info('getting bucket {}'.format(
                                current_app.config.get('SWAG_BUCKET')))

        swag = SWAGManager(**parse_swag_config_options(current_app.config.get('SWAG_OPTS')))

        all_accounts = swag.get_all(current_app.config.get('SWAG_FILTER'))

        service_enabled_requirement = current_app.config.get('SWAG_SERVICE_ENABLED_REQUIREMENT', None)
        if service_enabled_requirement:
            all_accounts = swag.get_service_enabled(service_enabled_requirement, accounts_list=all_accounts)

    except (KeyError, InvalidSWAGDataException, Exception) as e:
        current_app.logger.error('Account names passed but SWAG not configured or unavailable: {}'.format(e))

    if 'all' in account_names:
        return [account['id'] for account in all_accounts]

    lookup = {account['name']: Bunch(account) for account in all_accounts}

    for account in all_accounts:
        # get the right key, depending on whether we're using swag v1 or v2
        alias_key = 'aliases' if account['schemaVersion'] == '2' else 'alias'
        for alias in account[alias_key]:
            lookup[alias] = Bunch(account)

    for name in account_names:
        if name not in lookup:
            current_app.logger.warn('Could not find an account named %s'
                                    % name)
            continue

        account_number = lookup[name].get('id', None)
        if account_number:
            matching_accounts.append(account_number)

    return matching_accounts


class GunicornServer(Command):
    """
    This is the main GunicornServer server, it runs the flask app with gunicorn and
    uses any configuration options passed to it.
    You can pass all standard gunicorn flags to this command as if you were
    running gunicorn itself.
    For example:
    aardvark start_api -w 4 -b 127.0.0.0:8002
    Will start gunicorn with 4 workers bound to 127.0.0.0:8002
    """
    description = 'Run the app within Gunicorn'

    def get_options(self):
        options = []
        try:
            from gunicorn.config import make_settings
        except ImportError:
            # Gunicorn does not yet support Windows.
            # See issue #524. https://github.com/benoitc/gunicorn/issues/524
            # For dev on Windows, make this an optional import.
            print('Could not import gunicorn, skipping.')
            return options

        settings = make_settings()
        for setting, klass in settings.items():
            if klass.cli:
                if klass.action:
                    if klass.action == 'store_const':
                        options.append(Option(*klass.cli, const=klass.const, action=klass.action))
                    else:
                        options.append(Option(*klass.cli, action=klass.action))
                else:
                    options.append(Option(*klass.cli))
        return options

    def run(self, *args, **kwargs):
        from gunicorn.app.wsgiapp import WSGIApplication

        app = WSGIApplication()

        app.app_uri = 'aardvark:create_app()'
        return app.run()


def main():
    manager.add_command("start_api", GunicornServer())
    manager.run()


if __name__ == '__main__':
    manager.add_command("start_api", GunicornServer())
    manager.run()



#ensure absolute import for python3
from __future__ import absolute_import

import better_exceptions  # noqa
import datetime
import json

from flask import abort, jsonify
from flask import Blueprint
from flask_restful import Api, Resource, reqparse
from flask import Flask
import sqlalchemy as sa

from aardvark.model import AWSIAMObject


mod = Blueprint('advisor', __name__)
api = Api(mod)
app = Flask(__name__)


class RoleSearch(Resource):
    """
    Search for roles by phrase, regex, or by ARN.
    """
    def __init__(self):
        super(RoleSearch, self).__init__()
        self.reqparse = reqparse.RequestParser()

    def combine(self, aa):
        del aa['count']
        del aa['page']
        del aa['total']

        usage = dict()
        for arn, services in aa.items():
            for service in services:
                namespace = service.get('serviceNamespace')
                last_authenticated = service.get('lastAuthenticated')
                if namespace not in usage:
                    usage[namespace] = service
                else:
                    count_entities = usage[namespace]['totalAuthenticatedEntities'] + service['totalAuthenticatedEntities']
                    if last_authenticated > usage[namespace]['lastAuthenticated']:
                        usage[namespace] = service
                    usage[namespace]['totalAuthenticatedEntities'] = count_entities

        for namespace, service in usage.items():
            last_authenticated = service['lastAuthenticated']
            dt_last_authenticated = datetime.datetime.fromtimestamp(last_authenticated / 1e3)
            dt_starting = datetime.datetime.utcnow() - datetime.timedelta(days=90)
            usage[namespace]['USED_LAST_90_DAYS'] = dt_last_authenticated > dt_starting

        return jsonify(usage)

    # undocumented convenience pass-through so we can query directly from browser
    @app.route('/advisors')
    def get(self):
        return(self.post())

    @app.route('/advisors')
    def post(self):
        """Get access advisor data for role(s)
        Returns access advisor information for role(s) that match filters
        ---
        consumes:
          - 'application/json'
        produces:
          - 'application/json'

        parameters:
          - name: page
            in: query
            type: integer
            description: return results from given page of total results
            required: false
          - name: count
            in: query
            type: integer
            description: specifies how many results should be return per page
            required: false
          - name: combine
            in: query
            type: boolean
            description: combine access advisor data for all results [Default False]
            required: false
          - name: query
            in: body
            schema:
              $ref: '#/definitions/QueryBody'
            description: |
                one or more query parameters in a JSON blob.  Filter
                parameters build on eachother.

                Options are:

                1) arn list - a list of one or more specific arns

                2) phrase matching - search for ARNs like the one supplied

                3) regex - match a supplied regular expression.

        definitions:
          AdvisorData:
            type: object
            properties:
              lastAuthenticated:
                type: number
              lastAuthenticatedEntity:
                type: string
              lastUpdated:
                type: string
              serviceName:
                type: string
              serviceNamespace:
                type: string
              totalAuthenticatedEntities:
                type: number
          QueryBody:
            type: object
            properties:
              phrase:
                type: string
              regex:
                type: string
              arn:
                type: array
                items: string
          Results:
            type: array
            items:
              $ref: '#/definitions/AdvisorData'

        responses:
          200:
            description: Query successful, results in body
            schema:
              $ref: '#/definitions/AdvisorData'
          400:
            description: Bad request - error message in body
        """
        self.reqparse.add_argument('page', type=int, default=1)
        self.reqparse.add_argument('count', type=int, default=30)
        self.reqparse.add_argument('combine', type=str, default='false')
        self.reqparse.add_argument('phrase', default=None)
        self.reqparse.add_argument('regex', default=None)
        self.reqparse.add_argument('arn', default=None, action='append')
        try:
            args = self.reqparse.parse_args()
        except Exception as e:
            abort(400, str(e))

        page = args.pop('page')
        count = args.pop('count')
        combine = args.pop('combine', 'false')
        combine = combine.lower() == 'true'
        phrase = args.pop('phrase', '')
        arns = args.pop('arn', [])
        regex = args.pop('regex', '')
        items = None

        # default unfiltered query
        query = AWSIAMObject.query

        try:
            if phrase:
                query = query.filter(AWSIAMObject.arn.ilike('%' + phrase + '%'))

            if arns:
                query = query.filter(
                    sa.func.lower(AWSIAMObject.arn).in_([arn.lower() for arn in arns]))

            if regex:
                query = query.filter(AWSIAMObject.arn.regexp(regex))

            items = query.paginate(page, count)
        except Exception as e:
            abort(400, str(e))

        if not items:
            items = AWSIAMObject.query.paginate(page, count)

        values = dict(page=items.page, total=items.total, count=len(items.items))
        for item in items.items:
            item_values = []
            for advisor_data in item.usage:
                item_values.append(dict(
                    lastAuthenticated=advisor_data.lastAuthenticated,
                    serviceName=advisor_data.serviceName,
                    serviceNamespace=advisor_data.serviceNamespace,
                    lastAuthenticatedEntity=advisor_data.lastAuthenticatedEntity,
                    totalAuthenticatedEntities=advisor_data.totalAuthenticatedEntities,
                    lastUpdated=item.lastUpdated
                ))
            values[item.arn] = item_values

        if combine and items.total > len(items.items):
            abort(400, "Error: Please specify a count of at least {}.".format(items.total))
        elif combine:
            return self.combine(values)

        return jsonify(values)


api.add_resource(RoleSearch, '/advisors')



#ensure absolute import for python3
from __future__ import absolute_import

import copy
import time

from blinker import Signal
from cloudaux.aws.iam import list_roles, list_users
from cloudaux.aws.sts import boto3_cached_conn
from cloudaux.aws.decorators import rate_limited


class AccountToUpdate(object):
    on_ready = Signal()
    on_complete = Signal()
    on_error = Signal()
    on_failure = Signal()

    def __init__(self, current_app, account_number, role_name, arns_list):
        self.current_app = current_app
        self.account_number = account_number
        self.role_name = role_name
        self.arn_list = arns_list
        self.conn_details = {
            'account_number': account_number,
            'assume_role': role_name,
            'session_name': 'aardvark',
            'region': self.current_app.config.get('REGION') or 'us-east-1',
            'arn_partition': self.current_app.config.get('ARN_PARTITION') or 'aws'
        }
        self.max_access_advisor_job_wait = 5 * 60  # Wait 5 minutes before giving up on jobs

    def update_account(self):
        """
        Updates Access Advisor data for a given AWS account.
        1) Gets list of IAM Role ARNs in target account.
        2) Gets IAM credentials in target account.
        3) Calls GenerateServiceLastAccessedDetails for each role
        4) Calls GetServiceLastAccessedDetails for each role to retrieve data

        :return: Return code and JSON Access Advisor data for given account
        """
        self.on_ready.send(self)
        arns = self._get_arns()

        if not arns:
            self.current_app.logger.warn("Zero ARNs collected. Exiting")
            exit(-1)

        client = self._get_client()
        try:
            details = self._call_access_advisor(client, list(arns))
        except Exception as e:
            self.on_failure.send(self, error=e)
            self.current_app.logger.exception('Failed to call access advisor', exc_info=True)
            return 255, None
        else:
            self.on_complete.send(self)
            return 0, details

    def _get_arns(self):
        """
        Gets a list of all Role ARNs in a given account, optionally limited by
        class property ARN filter
        :return: list of role ARNs
        """
        client = boto3_cached_conn(
            'iam', service_type='client', **self.conn_details)

        account_arns = set()

        for role in list_roles(**self.conn_details):
            account_arns.add(role['Arn'])

        for user in list_users(**self.conn_details):
            account_arns.add(user['Arn'])

        for page in client.get_paginator('list_policies').paginate(Scope='Local'):
            for policy in page['Policies']:
                account_arns.add(policy['Arn'])

        for page in client.get_paginator('list_groups').paginate():
            for group in page['Groups']:
                account_arns.add(group['Arn'])

        result_arns = set()
        for arn in self.arn_list:
            if arn.lower() == 'all':
                return account_arns

            if arn not in account_arns:
                self.current_app.logger.warn("Provided ARN {arn} not found in account.".format(arn=arn))
                continue

            result_arns.add(arn)

        return list(result_arns)

    def _get_client(self):
        """
        Assumes into the target account and obtains IAM client

        :return: boto3 IAM client in target account & role
        """
        client = boto3_cached_conn(
            'iam', **self.conn_details)
        return client

    def _call_access_advisor(self, iam, arns):
        jobs = self._generate_job_ids(iam, arns)
        details = self._get_job_results(iam, jobs)
        if arns and not details:
            self.current_app.error("Didn't get any results from Access Advisor")
        return details

    @rate_limited()
    def _generate_service_last_accessed_details(self, iam, arn):
        """ Wrapping the actual AWS API calls for rate limiting protection. """
        return iam.generate_service_last_accessed_details(Arn=arn)['JobId']

    @rate_limited()
    def _get_service_last_accessed_details(self, iam, job_id):
        """ Wrapping the actual AWS API calls for rate limiting protection. """
        return iam.get_service_last_accessed_details(JobId=job_id)

    def _generate_job_ids(self, iam, arns):
        jobs = {}
        for role_arn in arns:
            try:
                job_id = self._generate_service_last_accessed_details(iam, role_arn)
                jobs[job_id] = role_arn
            except iam.exceptions.NoSuchEntityException:
                """ We're here because this ARN disappeared since the call to self._get_arns(). Log the missing ARN and move along.  """
                self.current_app.logger.info('ARN {arn} found gone when fetching details'.format(arn=role_arn))
            except Exception as e:
                self.on_error.send(self, error=e)
                self.current_app.logger.error('Could not gather data from {0}.'.format(role_arn), exc_info=True)
        return jobs

    def _get_job_results(self, iam, jobs):
        access_details = {}
        job_queue = list(jobs.keys())
        last_job_completion_time = time.time()

        while job_queue:

            # Check for timeout
            now = time.time()
            if now - last_job_completion_time > self.max_access_advisor_job_wait:
                # We ran out of time, some jobs are unfinished
                self._log_unfinished_jobs(job_queue, jobs)
                break

            # Pull next job ID
            job_id = job_queue.pop()
            role_arn = jobs[job_id]
            try:
                details = self._get_service_last_accessed_details(iam, job_id)
            except Exception as e:
                self.on_error.send(self, error=e)
                self.current_app.logger.error('Could not gather data from {0}.'.format(role_arn), exc_info=True)
                continue

            # Check job status
            if details['JobStatus'] == 'IN_PROGRESS':
                job_queue.append(job_id)
                if not job_queue:  # We're hanging on the last job, let's hang back for a bit
                    time.sleep(1)
                continue

            # Check for job failure
            if details['JobStatus'] != 'COMPLETED':
                log_str = "Job {job_id} finished with unexpected status {status} for ARN {arn}.".format(
                    job_id=job_id,
                    status=details['JobStatus'],
                    arn=role_arn)

                failing_arns = self.current_app.config.get('FAILING_ARNS', {})
                if role_arn in failing_arns:
                    self.current_app.logger.info(log_str)
                else:
                    self.current_app.logger.error(log_str)

                continue

            # Job status must be COMPLETED. Save result.
            last_job_completion_time = time.time()
            updated_list = []

            for detail in details.get('ServicesLastAccessed'):
                # create a copy, we're going to modify the time to epoch
                updated_item = copy.copy(detail)

                # AWS gives a datetime, convert to epoch
                last_auth = detail.get('LastAuthenticated')
                if last_auth:
                    last_auth = int(time.mktime(last_auth.timetuple()) * 1000)
                else:
                    last_auth = 0

                updated_item['LastAuthenticated'] = last_auth
                updated_list.append(updated_item)

            access_details[role_arn] = updated_list

        return access_details

    def _log_unfinished_jobs(self, job_queue, job_details):
        for job_id in job_queue:
            role_arn = job_details[job_id]
            self.current_app.logger.error("Job {job_id} for ARN {arn} didn't finish".format(
                job_id=job_id,
                arn=role_arn,
            ))



'''Test cases for the manage.config() function.

We test the following for the configurable parameters:
- Exactly the expected parameters appear in the config file.
- The expected parameters have the expected values in the config file.


In the event of test failures it may be helpful to examine the command
used to launch aardvark config, any responses to prompts and the
resulting configuration file. To this end, these artifacts will be
saved in locations controlled by the environment variable
SWAG_CONFIG_TEST_ARCHIVE_DIR. Set this environment variable to the
absolute path to the directory to use to archive test artifacts. The
default is '/tmp'.

By default, artifacts will only be saved in the event of test
failures. To force archiving of test artifacts regardless of test
status, set the SWAG_CONFIG_TEST_ALWAYS_ARCHIVE to 1 (or anything
truthy).

Archiving occurs in tearDown(). The config file, the command line call
to "aardvark config" and a transcript of the command line interaction
will be saved in the archive directory in files named "command.*" and
"config.py.*". The wildcard will be replaced by default with the name
of the "test_*" function that was running when the failure occcured.


The command lines used to call aardvark config in each test case will
always be archived. Set the SWAG_CONFIG_TEST_COMMAND_ARCHIVE_DIR
environment variable to the absolute path to the directory to use to
archive these commands.  The default is to use the same value as
SWAG_CONFIG_TEST_ARCHIVE_DIR. This will create a record of all the
command lines executed in execution of these test cases, in files
names "commands.[TestClassName]".

'''

#adding for py3 support
from __future__ import absolute_import

import inspect
import os
import shutil
import tempfile

import unittest

from aardvark import manage
import pexpect

# These are fast command line script interactions, eight seconds is forever.
EXPECT_TIMEOUT = 8

CONFIG_FILENAME = 'config.py'

ALWAYS_ARCHIVE = os.environ.get('SWAG_CONFIG_TEST_ALWAYS_ARCHIVE')

# Locations where we will archive test artifacts.
DEFAULT_ARTIFACT_ARCHIVE_DIR = '/tmp'
ARTIFACT_ARCHIVE_DIR = (
    os.environ.get('SWAG_CONFIG_TEST_ARCHIVE_DIR') or
    DEFAULT_ARTIFACT_ARCHIVE_DIR
    )
COMMAND_ARCHIVE_DIR = (
    os.environ.get('SWAG_CONFIG_TEST_COMMAND_ARCHIVE_DIR') or
    ARTIFACT_ARCHIVE_DIR
    )

DEFAULT_LOCALDB_FILENAME = 'aardvark.db'
DEFAULT_AARDVARK_ROLE = 'Aardvark'
DEFAULT_NUM_THREADS = 5

# Specification of option names, default values, methods of extracting
# from config file, etc. The keys here are what we use as the 'handle'
# for each configurable option throughout this test file.
CONFIG_OPTIONS = {
    'swag_bucket': {
        'short': '-b',
        'long': '--swag-bucket',
        'config_key': 'SWAG_OPTS',
        'config_prompt': r'(?i).*SWAG.*BUCKET.*:',
        'getval': lambda x: x.get('swag.bucket_name') if x else None,
        'default': manage.DEFAULT_SWAG_BUCKET
        },
    'aardvark_role': {
        'short': '-a',
        'long': '--aardvark-role',
        'config_key': 'ROLENAME',
        'config_prompt': r'(?i).*ROLE.*NAME.*:',
        'getval': lambda x: x,
        'default': manage.DEFAULT_AARDVARK_ROLE
        },
    'db_uri': {
        'short': '-d',
        'long': '--db-uri',
        'config_key': 'SQLALCHEMY_DATABASE_URI',
        'getval': lambda x: x,
        'config_prompt': r'(?i).*DATABASE.*URI.*:',
        'default': None  # need to be in tmpdir.
        },
    'num_threads': {
        'short': None,
        'long': '--num-threads',
        'config_key': 'NUM_THREADS',
        'config_prompt': r'(?i).*THREADS.*:',
        'getval': lambda x: x,
        'default': manage.DEFAULT_NUM_THREADS
        },
    }

# Syntax sugar for getting default parameters for each option. Note
# that we reset the db_uri value after we change the working directory
# in setUpClass().
DEFAULT_PARAMETERS = dict([
    (k, v['default']) for k, v in CONFIG_OPTIONS.items()
    ])
DEFAULT_PARAMETERS_NO_SWAG = dict(DEFAULT_PARAMETERS, **{'swag_bucket': None})

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Uncomment to show lower level logging statements.
# import logging
# logger = logging.getLogger()
# logger.setLevel(logging.DEBUG)
# shandler = logging.StreamHandler()
# shandler.setLevel(logging.INFO)  # Pick one.
# <!-- # shandler.setLevel(logging.DEBUG)  # Pick one. -->
# formatter = logging.Formatter(
#     '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
#     )
# shandler.setFormatter(formatter)
# logger.addHandler(shandler)


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def default_db_uri():
    '''Return the default db_uri value at runtime.'''
    return '{localdb}:///{path}/{filename}'.format(
        localdb=manage.LOCALDB,
        path=os.getcwd(),
        filename=manage.DEFAULT_LOCALDB_FILENAME
        )


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def get_config_option_string(cmdline_option_spec, short_flags=True):
    '''Construct the options string for a call to aardvark config.'''

    option_substrings = []

    for param, value in cmdline_option_spec.items():
        flag = (
            CONFIG_OPTIONS[param]['short']
            if short_flags and CONFIG_OPTIONS[param]['short']
            else CONFIG_OPTIONS[param]['long']
            )
        option_substrings.append('{} {}'.format(flag, value))

    return ' '.join(option_substrings)


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def load_configfile(cmdline_option_spec):
    '''Evaluate the config values for the fields in cmdline_option_spec.'''

    all_config = {}
    with open(CONFIG_FILENAME) as in_file:
        exec(in_file.read(), all_config)
    # print all_config.keys()
    found_config = dict([
        (k, v['getval'](all_config.get(v['config_key'])))
        for (k, v) in CONFIG_OPTIONS.items()
        ])
    return found_config


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def get_expected_config(option_spec):
    '''Return a dict with the values that should be set by a config file.'''

    include_swag = ('swag_bucket' in option_spec)
    default_parameters = (
        DEFAULT_PARAMETERS if include_swag else DEFAULT_PARAMETERS_NO_SWAG
        )
    expected_config = dict(default_parameters)
    expected_config.update(option_spec)

    return expected_config


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TestConfigBase(unittest.TestCase):
    '''Base class for config test cases.'''

    # Throughout, the dicts cmdline_option_spec and config_option_spec
    # are defined with keys matching the keys in CONFIG_SPEC and the
    # values defining the value for the corresponding parameter, to be
    # delivered via a command line parameter to 'aardvark config' or
    # via entry after the appropriate prompt interactively.

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @classmethod
    def setUpClass(cls):
        '''Test case class common fixture setup.'''

        cls.tmpdir = tempfile.mkdtemp()
        cls.original_working_dir = os.getcwd()
        os.chdir(cls.tmpdir)

        cls.commands_issued = []

        # These depend on the current working directory set above.
        CONFIG_OPTIONS['db_uri']['default'] = default_db_uri()
        DEFAULT_PARAMETERS['db_uri'] = CONFIG_OPTIONS['db_uri']['default']
        DEFAULT_PARAMETERS_NO_SWAG['db_uri'] = (
            CONFIG_OPTIONS['db_uri']['default']
            )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @classmethod
    def tearDownClass(cls):
        '''Test case class common fixture teardown.'''

        os.chdir(cls.original_working_dir)
        cls.clean_tmpdir()
        os.rmdir(cls.tmpdir)

        command_archive_filename = '.'.join(['commands', cls.__name__])
        command_archive_path = os.path.join(
            COMMAND_ARCHIVE_DIR, command_archive_filename
            )

        with open(command_archive_path, 'w') as fptr:
            fptr.write('\n'.join(cls.commands_issued) + '\n')

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    @classmethod
    def clean_tmpdir(cls):
        '''Remove all content from cls.tmpdir.'''
        for root, dirs, files in os.walk(cls.tmpdir, topdown=False):
            for name in files:
                os.remove(os.path.join(root, name))
            for name in dirs:
                os.rmdir(os.path.join(root, name))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def setUp(self):
        '''Test case common fixture setup.'''
        self.clean_tmpdir()
        self.assertFalse(os.path.exists(CONFIG_FILENAME))
        self.last_transcript = []
        self.archive_case_artifacts_as = None

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def tearDown(self):
        '''Test case common fixture teardown.'''

        # Archive the last command and config file created, if indicated.
        if self.archive_case_artifacts_as:

            command_archive_path = self.archive_path(
                'command', self.archive_case_artifacts_as
                )
            config_archive_path = self.archive_path(
                CONFIG_FILENAME, self.archive_case_artifacts_as
                )

            with open(command_archive_path, 'w') as fptr:
                fptr.write(self.last_config_command + '\n')
                if self.last_transcript:
                    fptr.write(
                        '\n'.join(
                            map(lambda x: str(x), self.last_transcript)
                            ) + '\n'
                        )

            if os.path.exists(CONFIG_FILENAME):
                shutil.copyfile(CONFIG_FILENAME, config_archive_path)
            else:
                with open(config_archive_path, 'w') as fptr:
                    fptr.write(
                        '(no {} file found in {})\n'.format(
                            CONFIG_FILENAME, os.getcwd()
                            )
                        )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def archive_path(self, filename, suffix):
        '''Return the path to an archive file.'''
        archive_filename = '.'.join([filename, suffix])
        archive_path = os.path.join(ARTIFACT_ARCHIVE_DIR, archive_filename)
        return archive_path

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def call_aardvark_config(
            self,
            cmdline_option_spec=None,
            input_option_spec=None,
            prompt=True,
            short_flags=False
            ):
        '''Call aardvark config and interact as necessary.'''

        cmdline_option_spec = cmdline_option_spec or {}
        input_option_spec = input_option_spec or {}

        command = 'aardvark config' + ('' if prompt else ' --no-prompt')
        self.last_config_command = '{} {}'.format(
            command,
            get_config_option_string(
                cmdline_option_spec, short_flags=short_flags
                )
            )

        self.commands_issued.append(self.last_config_command)
        spawn_config = pexpect.spawn(self.last_config_command)

        self.conduct_config_prompt_sequence(
            spawn_config, input_option_spec
            )

        # If we didn't wrap up the session, something's amiss.
        self.assertFalse(spawn_config.isalive())

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def conduct_config_prompt_sequence(self, spawned, input_option_spec):
        '''Carry out the steps in the config prompt sequence.'''

        # The order is all that tells us which of these match in a pexpect
        # call, so we can't use a dict here.
        control_prompts = [
            (pexpect.EOF, 'eof'),
            (pexpect.TIMEOUT, 'timeout')
            ]
        config_option_prompts = [
            (v['config_prompt'], k)
            for k, v in CONFIG_OPTIONS.items()
            ]
        expect_prompts = [
            (r'(?i).*Do you use SWAG.*:', 'use_swag'),
            ]

        expect_prompts.extend(config_option_prompts)
        expect_prompts.extend(control_prompts)

        response_spec = input_option_spec
        response_spec['use_swag'] = (
            'y' if 'swag_bucket' in input_option_spec else 'N'
            )

        while spawned.isalive():

            prompt_index = spawned.expect(
                [x[0] for x in expect_prompts], timeout=EXPECT_TIMEOUT
                )
            self.last_transcript.append(spawned.after)

            prompt_received = expect_prompts[prompt_index][1]
            if prompt_received in [x[1] for x in control_prompts]:
                return

            response = response_spec.get(prompt_received)
            response = '' if response is None else response
            spawned.sendline(str(response))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def case_worker(
            self,
            cmdline_option_spec=None,
            input_option_spec=None,
            prompt=False,
            short_flags=False,
            expect_config_file=True,
            archive_as=None
            ):
        '''Carry out common test steps.

        Parameters:

            cmdline_option_spec (dict or None):
                A dictionary specifying the options and values to pass
                to aardvark config as command line flags.

            input_option_spec (dict or None):
                A dictionary specifying the options and values to pass
                to aardvark config as prompted interactive input.

            prompt (bool, default: False):
                If False, set the --no-prompt option when calling
                aardvark config.

            short_flags (bool, default: False):
                If True, set the --short-flags  option when calling
                aardvark config.

            expect_config_file (bool, default: True):
                If True, test for the presence and correctness of a
                config file after calling aardvark config. If false,
                test for the absence of a config file.

            archive_as (str or None):
                The "unique" string to use when constructing a
                filename for archiving artifacts of this test case. If
                None, the name of the nearest calling function in the
                call stack named "test_*" will be used, if one can be
                found; otherwise a possibly non-unique string will be
                used.

        '''

        cmdline_option_spec = cmdline_option_spec or {}
        input_option_spec = input_option_spec or {}
        # Combined, for validation.
        option_spec = dict(cmdline_option_spec, **input_option_spec)

        if not archive_as:
            # Get the calling test case function's name, for
            # archiving. We'll take the first caller in the stack
            # whose name starts with 'test_'.
            caller_names = [
                inspect.getframeinfo(frame[0]).function
                for frame in inspect.stack()
                if inspect.getframeinfo(frame[0]).function.startswith('test_')
                ]
            archive_as = caller_names[0] if caller_names else 'unknown_test'

        # Turn on failed-case archive.
        self.archive_case_artifacts_as = archive_as

        self.assertFalse(os.path.exists(CONFIG_FILENAME))
        self.call_aardvark_config(
            cmdline_option_spec=cmdline_option_spec,
            input_option_spec=input_option_spec,
            prompt=prompt,
            short_flags=short_flags
            )

        if expect_config_file:
            self.assertTrue(os.path.exists(CONFIG_FILENAME))

            found_config = load_configfile(cmdline_option_spec)
            expected_config = get_expected_config(option_spec)

            self.assertCountEqual(expected_config.keys(), found_config.keys())
            for k, v in found_config.items():
                self.assertEqual((k, v), (k, expected_config[k]))

        else:
            self.assertFalse(os.path.exists(CONFIG_FILENAME))

        # Turn off failed-case archive unless we're forcing archiving.
        if not ALWAYS_ARCHIVE:
            self.archive_case_artifacts_as = None


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TestConfigNoPrompt(TestConfigBase):
    '''Test cases for config --no-prompt.'''

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def test_no_prompt_defaults(self):
        '''Test with no-prompt and all default arguments.'''

        self.case_worker()

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def test_no_prompt_all_parameters(self):
        '''Test with no-prompt and all parameters.'''

        cmdline_option_spec = {
            'swag_bucket': 'bucket_123',
            'aardvark_role': 'role_123',
            'db_uri': 'db_uri_123',
            'num_threads': 4
            }

        self.case_worker(
            cmdline_option_spec=cmdline_option_spec,
            )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def test_no_prompt_all_parameters_short(self):
        '''Test with no-prompt and short parameters.'''

        cmdline_option_spec = {
            'swag_bucket': 'bucket_123',
            'aardvark_role': 'role_123',
            'db_uri': 'db_uri_123',
            'num_threads': 4
            }

        self.case_worker(
            cmdline_option_spec=cmdline_option_spec,
            short_flags=True
            )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def test_no_prompt_no_swag(self):
        '''Test with no-prompt and all non-swag parameters.'''

        cmdline_option_spec = {
            'aardvark_role': 'role_123',
            'db_uri': 'db_uri_123',
            'num_threads': 4
            }

        self.case_worker(
            cmdline_option_spec=cmdline_option_spec,
            )


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TestConfigPrompt(TestConfigBase):
    '''Test cases for config with prompting.'''

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def test_prompted_defaults(self):
        '''Test with no parameters specified.'''

        self.case_worker(
            prompt=True
            )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def test_prompted_all_cmdline_parameters(self):
        '''Test with all parameters passed as options.'''

        cmdline_option_spec = {
            'swag_bucket': 'bucket_123',
            'aardvark_role': 'role_123',
            'db_uri': 'db_uri_123',
            'num_threads': 4
            }

        self.case_worker(
            cmdline_option_spec=cmdline_option_spec,
            prompt=True
            )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def test_prompted_no_swag(self):
        '''Test with all non-swag parameters interactively.'''

        input_option_spec = {
            'aardvark_role': 'role_123',
            'db_uri': 'db_uri_123',
            'num_threads': 4
            }

        self.case_worker(
            input_option_spec=input_option_spec,
            prompt=True
            )

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Define test suites.
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
load_case = unittest.TestLoader().loadTestsFromTestCase
all_suites = {
    'testconfignoprompt': load_case(TestConfigNoPrompt),
    'testconfigprompt': load_case(TestConfigPrompt)
    }

master_suite = unittest.TestSuite(all_suites.values())

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
if __name__ == '__main__':
    unittest.main()



#ensure absolute import for python3
from __future__ import absolute_import

import os.path
import logging
from logging import DEBUG, Formatter, StreamHandler
from logging.config import dictConfig
import sys

from flask_sqlalchemy import SQLAlchemy
from flask import Flask
from flasgger import Swagger

db = SQLAlchemy()

from aardvark.view import mod as advisor_bp  # noqa

BLUEPRINTS = [
    advisor_bp
]

API_VERSION = '1'


def create_app():
    app = Flask(__name__, static_url_path='/static')
    Swagger(app)

    path = _find_config()
    if not path:
        print('No config')
        app.config.from_pyfile('_config.py')
    else:
        app.config.from_pyfile(path)

    # For ELB and/or Eureka
    @app.route('/healthcheck')
    def healthcheck():
        """Healthcheck
        Simple healthcheck that indicates the services is up
        ---
        responses:
          200:
            description: service is up
        """
        return 'ok'

    # Blueprints
    for bp in BLUEPRINTS:
        app.register_blueprint(bp, url_prefix="/api/{0}".format(API_VERSION))

    # Extensions:
    db.init_app(app)
    setup_logging(app)

    return app


def _find_config():
    """Search for config.py in order of preference and return path if it exists, else None"""
    CONFIG_PATHS = [os.path.join(os.getcwd(), 'config.py'),
                    '/etc/aardvark/config.py',
                    '/apps/aardvark/config.py']
    for path in CONFIG_PATHS:
        if os.path.exists(path):
            return path
    return None


def setup_logging(app):
    if not app.debug:
        if app.config.get('LOG_CFG'):
            # initialize the Flask logger (removes all handlers)
            app.logger
            dictConfig(app.config.get('LOG_CFG'))
            app.logger = logging.getLogger(__name__)
        else:
            handler = StreamHandler(stream=sys.stderr)

            handler.setFormatter(Formatter(
                '%(asctime)s %(levelname)s: %(message)s '
                '[in %(pathname)s:%(lineno)d]'))
            app.logger.setLevel(app.config.get('LOG_LEVEL', DEBUG))
            app.logger.addHandler(handler)



#ensure absolute import for python3
from __future__ import absolute_import

import datetime

from flask import current_app
from sqlalchemy import BigInteger, Column, Integer, Text, TIMESTAMP
import sqlalchemy.exc
from sqlalchemy.orm import relationship
from sqlalchemy.schema import ForeignKey

from aardvark import db
from aardvark.utils.sqla_regex import String


class AWSIAMObject(db.Model):
    """
    Meant to model AWS IAM Object Access Advisor.
    """
    __tablename__ = "aws_iam_object"
    id = Column(Integer, primary_key=True)
    arn = Column(String(2048), nullable=True, index=True, unique=True)
    lastUpdated = Column(TIMESTAMP)
    usage = relationship("AdvisorData", backref="item", cascade="all, delete, delete-orphan",
                         foreign_keys="AdvisorData.item_id")

    @staticmethod
    def get_or_create(arn):
        item = AWSIAMObject.query.filter(AWSIAMObject.arn == arn).scalar()

        added = False
        try:
            item = AWSIAMObject.query.filter(AWSIAMObject.arn == arn).scalar()
        except sqlalchemy.exc.SQLAlchemyException as e:
            current_app.logger.error('Database exception: {}'.format(e.message))

        if not item:
            item = AWSIAMObject(arn=arn, lastUpdated=datetime.datetime.utcnow())
            added = True
        else:
            item.lastUpdated = datetime.datetime.utcnow()
        db.session.add(item)

        # we only need a refresh if the object was created
        if added:
            db.session.commit()
            db.session.refresh(item)
        return item


class AdvisorData(db.Model):
    """
    Models certain IAM Access Advisor Data fields.

    {
      "totalAuthenticatedEntities": 1,
      "lastAuthenticatedEntity": "arn:aws:iam::XXXXXXXX:role/name",
      "serviceName": "Amazon Simple Systems Manager",
      "lastAuthenticated": 1489176000000,
      "serviceNamespace": "ssm"
    }
    """
    __tablename__ = "advisor_data"
    id = Column(Integer, primary_key=True)
    item_id = Column(Integer, ForeignKey("aws_iam_object.id"), nullable=False, index=True)
    lastAuthenticated = Column(BigInteger)
    serviceName = Column(String(128), index=True)
    serviceNamespace = Column(String(64), index=True)
    lastAuthenticatedEntity = Column(Text)
    totalAuthenticatedEntities = Column(Integer)

    @staticmethod
    def create_or_update(item_id, lastAuthenticated, serviceName, serviceNamespace, lastAuthenticatedEntity,
                         totalAuthenticatedEntities):
        serviceName = serviceName[:128]
        serviceNamespace = serviceNamespace[:64]
        item = None
        try:
            item = AdvisorData.query.filter(AdvisorData.item_id == item_id).filter(AdvisorData.serviceNamespace ==
                                                                                   serviceNamespace).scalar()
        except sqlalchemy.exc.SQLAlchemyError as e:
            current_app.logger.error('Database error: {} item_id: {} serviceNamespace: {}'.format(e.args[0], item_id,
                                     serviceNamespace)) #exception.messsage not supported in py3 e.args[0] replacement

        if not item:
            item = AdvisorData(item_id=item_id,
                               lastAuthenticated=lastAuthenticated,
                               serviceName=serviceName,
                               serviceNamespace=serviceNamespace,
                               lastAuthenticatedEntity=lastAuthenticatedEntity,
                               totalAuthenticatedEntities=totalAuthenticatedEntities)
            db.session.add(item)
            return

        if lastAuthenticated > item.lastAuthenticated:
            item.lastAuthenticated = lastAuthenticated
            db.session.add(item)

        elif lastAuthenticated < item.lastAuthenticated:
            """
            lastAuthenticated is obtained by calling get_service_last_accessed_details() method of the boto3 iam client.
            When there is no AA data about a service, the lastAuthenticated key is missing from the returned dictionary.
            This is perfectly valid, either because the service in question was not accessed in the past 365 days or
            the entity granting  access to it was created recently enough that no AA data is available yet (it can take up to
            4 hours for this to happen).
            When this happens, the AccountToUpdate._get_job_results() method will set lastAuthenticated to 0.
            Usually we don't want to persist such an entity, with one exception: there's already a recorded, non-zero lastAuthenticated
            timestamp persisted for this item. That means the service was accessed at some point in time, but now more than 365 passed since
            the last access, so AA no longer returns a timestamp for it.
            """
            if lastAuthenticated == 0:
                current_app.logger.warn('Previously seen object not accessed in the past 365 days '
                                        '(got null lastAuthenticated from AA). Setting to 0. '
                                        'Object {} service {} previous timestamp {}'.format(item.item_id, item.serviceName, item.lastAuthenticated))
                item.lastAuthenticated = 0
                db.session.add(item)
            else:
                current_app.logger.error("Received an older time than previously seen for object {} service {} ({la} < {ila})!".format(item.item_id,
                                                                                                                                       item.serviceName,
                                                                                                                                       la=lastAuthenticated,
                                                                                                                                       ila=item.lastAuthenticated))
